# validated: 2018-01-14 DV a68a0c3ebf0b libraries/driver/include/ctre/phoenix/Sensors/PigeonIMU_StickyFaults.h
from .faultsbase import FaultsBase

__all__ = ['PigeonIMU_StickyFaults']


class PigeonIMU_StickyFaults(FaultsBase):
    pass
