# validated: 2018-01-14 DV a68a0c3ebf0b libraries/driver/include/ctre/phoenix/CANifierStickyFaults.h
from .faultsbase import FaultsBase

__all__ = ['CANifierStickyFaults']


class CANifierStickyFaults(FaultsBase):
    pass
